import json
import sys
import uuid
from http import HTTPStatus
from pathlib import Path

from fastapi import APIRouter, Depends, HTTPException, Response
from fastapi.responses import FileResponse
from fps.hooks import register_router  # type: ignore
from fps_auth_base import User, current_user, websocket_auth  # type: ignore
from fps_frontend.config import get_frontend_config  # type: ignore
from fps_yjs.routes import YDocWebSocketHandler  # type: ignore
from starlette.requests import Request  # type: ignore

from .config import get_kernel_config
from .kernel_driver.driver import KernelDriver  # type: ignore
from .kernel_driver.kernelspec import find_kernelspec, kernelspec_dirs
from .kernel_server.server import (  # type: ignore
    AcceptedWebSocket,
    KernelServer,
    kernels,
)
from .models import CreateSession, Execution, Session

router = APIRouter()

kernelspecs: dict = {}
sessions: dict = {}
prefix_dir: Path = Path(sys.prefix)


@router.on_event("shutdown")
async def stop_kernels():
    for kernel in kernels.values():
        await kernel["server"].stop()


@router.get("/api/kernelspecs")
async def get_kernelspecs(
    frontend_config=Depends(get_frontend_config),
    kernel_config=Depends(get_kernel_config),
    user: User = Depends(current_user(permissions={"kernelspecs": ["read"]})),
):
    for search_path in kernelspec_dirs():
        for path in Path(search_path).glob("*/kernel.json"):
            with open(path) as f:
                spec = json.load(f)
            name = path.parent.name
            resources = {
                f.stem: f"{frontend_config.base_url}kernelspecs/{name}/{f.name}"
                for f in path.parent.iterdir()
                if f.is_file() and f.name != "kernel.json"
            }
            kernelspecs[name] = {"name": name, "spec": spec, "resources": resources}
    return {"default": kernel_config.default_kernel, "kernelspecs": kernelspecs}


@router.get("/kernelspecs/{kernel_name}/{file_name}")
async def get_kernelspec(
    kernel_name,
    file_name,
    user: User = Depends(current_user()),
):
    for search_path in kernelspec_dirs():
        file_path = Path(search_path) / kernel_name / file_name
        if file_path.exists():
            return FileResponse(file_path)
    raise HTTPException(status_code=404, detail=f"Kernelspec {kernel_name}/{file_name} not found")


@router.get("/api/kernels")
async def get_kernels(
    user: User = Depends(current_user(permissions={"kernels": ["read"]})),
):
    results = []
    for kernel_id, kernel in kernels.items():
        results.append(
            {
                "id": kernel_id,
                "name": kernel["name"],
                "connections": kernel["server"].connections,
                "last_activity": kernel["server"].last_activity["date"],
                "execution_state": kernel["server"].last_activity["execution_state"],
            }
        )
    return results


@router.delete("/api/sessions/{session_id}", status_code=204)
async def delete_session(
    session_id: str,
    user: User = Depends(current_user(permissions={"sessions": ["write"]})),
):
    kernel_id = sessions[session_id]["kernel"]["id"]
    kernel_server = kernels[kernel_id]["server"]
    await kernel_server.stop()
    del kernels[kernel_id]
    del sessions[session_id]
    return Response(status_code=HTTPStatus.NO_CONTENT.value)


@router.patch("/api/sessions/{session_id}")
async def rename_session(
    request: Request,
    user: User = Depends(current_user(permissions={"sessions": ["write"]})),
):
    rename_session = await request.json()
    session_id = rename_session.pop("id")
    for key, value in rename_session.items():
        sessions[session_id][key] = value
    return Session(**sessions[session_id])


@router.get("/api/sessions")
async def get_sessions(
    user: User = Depends(current_user(permissions={"sessions": ["read"]})),
):
    for session in sessions.values():
        kernel_id = session["kernel"]["id"]
        kernel_server = kernels[kernel_id]["server"]
        session["kernel"]["last_activity"] = kernel_server.last_activity["date"]
        session["kernel"]["execution_state"] = kernel_server.last_activity["execution_state"]
    return list(sessions.values())


@router.post(
    "/api/sessions",
    status_code=201,
    response_model=Session,
)
async def create_session(
    request: Request,
    user: User = Depends(current_user(permissions={"sessions": ["write"]})),
):
    create_session = CreateSession(**(await request.json()))
    kernel_name = create_session.kernel.name
    kernel_server = KernelServer(
        kernelspec_path=Path(find_kernelspec(kernel_name)).as_posix(),
        kernel_cwd=str(Path(create_session.path).parent),
    )
    kernel_id = str(uuid.uuid4())
    kernels[kernel_id] = {"name": kernel_name, "server": kernel_server, "driver": None}
    await kernel_server.start()
    session_id = str(uuid.uuid4())
    session = {
        "id": session_id,
        "path": create_session.path,
        "name": create_session.name,
        "type": create_session.type,
        "kernel": {
            "id": kernel_id,
            "name": create_session.kernel.name,
            "connections": kernel_server.connections,
            "last_activity": kernel_server.last_activity["date"],
            "execution_state": kernel_server.last_activity["execution_state"],
        },
        "notebook": {"path": create_session.path, "name": create_session.name},
    }
    sessions[session_id] = session
    return Session(**session)


@router.post("/api/kernels/{kernel_id}/restart")
async def restart_kernel(
    kernel_id,
    user: User = Depends(current_user(permissions={"kernels": ["write"]})),
):
    if kernel_id in kernels:
        kernel = kernels[kernel_id]
        await kernel["server"].restart()
        result = {
            "id": kernel_id,
            "name": kernel["name"],
            "connections": kernel["server"].connections,
            "last_activity": kernel["server"].last_activity["date"],
            "execution_state": kernel["server"].last_activity["execution_state"],
        }
        return result


@router.post("/api/kernels/{kernel_id}/execute")
async def execute_cell(
    request: Request,
    kernel_id,
    user: User = Depends(current_user(permissions={"kernels": ["write"]})),
):
    r = await request.json()
    execution = Execution(**r)
    if kernel_id in kernels:
        ynotebook = YDocWebSocketHandler.websocket_server.get_room(execution.document_id).document
        cell = ynotebook.get_cell(execution.cell_idx)
        cell["outputs"] = []

        kernel = kernels[kernel_id]
        if not kernel["driver"]:
            kernel["driver"] = driver = KernelDriver(
                kernelspec_path=Path(find_kernelspec(kernel["name"])).as_posix(),
                write_connection_file=False,
                connection_file=kernel["server"].connection_file_path,
            )
            await driver.connect()
        driver = kernel["driver"]

        await driver.execute(cell)
        ynotebook.set_cell(execution.cell_idx, cell)


@router.get("/api/kernels/{kernel_id}")
async def get_kernel(
    kernel_id,
    user: User = Depends(current_user(permissions={"kernels": ["read"]})),
):
    if kernel_id in kernels:
        kernel = kernels[kernel_id]
        result = {
            "id": kernel_id,
            "name": kernel["name"],
            "connections": kernel["server"].connections,
            "last_activity": kernel["server"].last_activity["date"],
            "execution_state": kernel["server"].last_activity["execution_state"],
        }
        return result


@router.websocket("/api/kernels/{kernel_id}/channels")
async def kernel_channels(
    kernel_id,
    session_id,
    websocket_permissions=Depends(websocket_auth(permissions={"kernels": ["execute"]})),
):
    if websocket_permissions is None:
        return
    websocket, permissions = websocket_permissions
    subprotocol = (
        "v1.kernel.websocket.jupyter.org"
        if "v1.kernel.websocket.jupyter.org" in websocket["subprotocols"]
        else None
    )
    await websocket.accept(subprotocol=subprotocol)
    accepted_websocket = AcceptedWebSocket(websocket, subprotocol)
    if kernel_id in kernels:
        kernel_server = kernels[kernel_id]["server"]
        await kernel_server.serve(accepted_websocket, session_id, permissions)


r = register_router(router)



import tempfile
from pathlib import Path

import nbconvert  # type: ignore
from fastapi import APIRouter, Depends
from fastapi.responses import FileResponse
from fps.hooks import register_router  # type: ignore
from fps_auth_base import User, current_user  # type: ignore

router = APIRouter()


@router.get("/api/nbconvert")
async def get_nbconvert_formats():
    return {
        name: {"output_mimetype": nbconvert.exporters.get_exporter(name).output_mimetype}
        for name in nbconvert.exporters.get_export_names()
    }


@router.get("/nbconvert/{format}/{path}")
async def get_nbconvert_document(
    format: str,
    path: str,
    download: bool,
    user: User = Depends(current_user(permissions={"nbconvert": ["read"]})),
):
    exporter = nbconvert.exporters.get_exporter(format)
    if download:
        media_type = "application/octet-stream"
    else:
        media_type = exporter.output_mimetype
    tmp_dir = Path(tempfile.mkdtemp())
    tmp_path = tmp_dir / (Path(path).stem + exporter().file_extension)
    with open(tmp_path, "wt") as f:
        f.write(exporter().from_filename(path)[0])
    return FileResponse(tmp_path, media_type=media_type, filename=tmp_path.name)


r = register_router(router)



from typing import Dict, List, Optional


def create_content(
    content: Optional[List],
    type: str,
    size: Optional[int],
    mimetype: Optional[str],
    name: str,
    path: str,
    format: Optional[str],
) -> Dict:
    return {
        "content": content,
        "created": None,
        "format": format,
        "last_modified": None,
        "mimetype": mimetype,
        "name": name,
        "path": path,
        "size": size,
        "type": type,
        "writable": True,
    }


def clear_content_values(content: Dict, keys: List[str] = []):
    for k in content:
        if k in keys:
            content[k] = None
        if k == "content" and isinstance(content[k], list):
            for c in content[k]:
                clear_content_values(c, keys)
    return content


def sort_content_by_name(content: Dict):
    for k in content:
        if k == "content" and isinstance(content[k], list):
            # FIXME: this sorting algorithm is terrible!
            names = [c["name"] for c in content[k]]
            names.sort()
            new_content = []
            for name in names:
                for i, c in enumerate(content[k]):
                    if c["name"] == name:
                        break
                content[k].pop(i)
                new_content.append(c)
            content[k] = new_content
            for c in content[k]:
                sort_content_by_name(c)
    return content



from fps.config import PluginModel, get_config  # type: ignore
from fps.hooks import register_config  # type: ignore


class KernelConfig(PluginModel):
    default_kernel: str = "python3"


def get_kernel_config():
    return get_config(KernelConfig)


c = register_config(KernelConfig)



import os
import socket
import subprocess
import time
from pathlib import Path

import pytest

pytest_plugins = (
    "fps.testing.fixtures",
    "fps_auth.fixtures",
)


@pytest.fixture()
def cwd():
    return Path(__file__).parents[1]


def get_open_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = str(s.getsockname()[1])
    s.close()
    return port


@pytest.fixture()
def start_jupyverse(auth_mode, clear_users, cwd, capfd):
    os.chdir(cwd)
    port = get_open_port()
    command_list = [
        "jupyverse",
        "--no-open-browser",
        f"--auth.mode={auth_mode}",
        "--auth.clear_users=" + str(clear_users).lower(),
        f"--port={port}",
    ]
    print(" ".join(command_list))
    p = subprocess.Popen(command_list)
    dtime, ttime, timeout = 0.1, 0, 10
    while True:
        time.sleep(dtime)
        ttime += dtime
        if ttime >= timeout:
            raise RuntimeError("Timeout while launching Jupyverse")
        out, err = capfd.readouterr()
        if "Application startup complete" in err:
            break
    url = f"http://127.0.0.1:{port}"
    yield url
    p.kill()



import asyncio
import json
import os
import signal
import uuid
from datetime import datetime
from typing import Dict, Iterable, List, Optional, cast

from fastapi import WebSocket, WebSocketDisconnect  # type: ignore
from starlette.websockets import WebSocketState

from ..kernel_driver.connect import (
    cfg_t,
    connect_channel,
    launch_kernel,
    read_connection_file,
)
from ..kernel_driver.connect import (
    write_connection_file as _write_connection_file,  # type: ignore
)
from ..kernel_driver.message import create_message, receive_message, send_message
from .message import (  # type: ignore
    deserialize_msg_from_ws_v1,
    from_binary,
    get_msg_from_parts,
    get_parent_header,
    get_zmq_parts,
    send_raw_message,
    serialize_msg_to_ws_v1,
    to_binary,
)

kernels: dict = {}


class AcceptedWebSocket:
    _websocket: WebSocket
    _accepted_subprotocol: Optional[str]

    def __init__(self, websocket, accepted_subprotocol):
        self._websocket = websocket
        self._accepted_subprotocol = accepted_subprotocol

    @property
    def websocket(self):
        return self._websocket

    @property
    def accepted_subprotocol(self):
        return self._accepted_subprotocol


class KernelServer:
    def __init__(
        self,
        kernelspec_path: str = "",
        kernel_cwd: str = "",
        connection_cfg: Optional[cfg_t] = None,
        connection_file: str = "",
        write_connection_file: bool = True,
        capture_kernel_output: bool = True,
    ) -> None:
        self.capture_kernel_output = capture_kernel_output
        self.kernelspec_path = kernelspec_path
        self.kernel_cwd = kernel_cwd
        self.connection_cfg = connection_cfg
        self.connection_file = connection_file
        self.write_connection_file = write_connection_file
        self.channel_tasks: List[asyncio.Task] = []
        self.sessions: Dict[str, AcceptedWebSocket] = {}
        # blocked messages and allowed messages are mutually exclusive
        self.blocked_messages: List[str] = []
        self.allowed_messages: Optional[List[str]] = None  # when None, all messages are allowed
        # when [], no message is allowed
        self.setup_connection_file()

    def setup_connection_file(self):
        if self.write_connection_file:
            self.connection_file_path, self.connection_cfg = _write_connection_file(
                self.connection_file
            )
        elif self.connection_file:
            self.connection_file_path = self.connection_file
            self.connection_cfg = read_connection_file(self.connection_file)
        else:
            if self.connection_cfg is None:
                raise RuntimeError("No connection_cfg")
        self.key = cast(str, self.connection_cfg["key"])

    def block_messages(self, message_types: Iterable[str] = []):
        # if using blocked messages, discard allowed messages
        self.allowed_messages = None
        if isinstance(message_types, str):
            message_types = [message_types]
        self.blocked_messages = list(message_types)

    def allow_messages(self, message_types: Optional[Iterable[str]] = None):
        # if using allowed messages, discard blocked messages
        self.blocked_messages = []
        if message_types is None:
            self.allowed_messages = None
            return
        if isinstance(message_types, str):
            message_types = [message_types]
        self.allowed_messages = list(message_types)

    @property
    def connections(self) -> int:
        return len(self.sessions)

    async def start(self) -> None:
        if not self.kernelspec_path:
            raise RuntimeError("Could not find a kernel, maybe you forgot to install one?")
        self.last_activity = {
            "date": datetime.utcnow().isoformat() + "Z",
            "execution_state": "starting",
        }
        self.kernel_process = await launch_kernel(
            self.kernelspec_path,
            self.connection_file_path,
            self.kernel_cwd,
            self.capture_kernel_output,
        )
        assert self.connection_cfg is not None
        identity = uuid.uuid4().hex.encode("ascii")
        self.shell_channel = connect_channel("shell", self.connection_cfg, identity=identity)
        self.stdin_channel = connect_channel("stdin", self.connection_cfg, identity=identity)
        self.control_channel = connect_channel("control", self.connection_cfg, identity=identity)
        self.iopub_channel = connect_channel("iopub", self.connection_cfg)
        await self._wait_for_ready()
        self.channel_tasks += [
            asyncio.create_task(self.listen("shell")),
            asyncio.create_task(self.listen("stdin")),
            asyncio.create_task(self.listen("control")),
            asyncio.create_task(self.listen("iopub")),
        ]

    async def stop(self) -> None:
        # FIXME: stop kernel in a better way
        try:
            self.kernel_process.send_signal(signal.SIGINT)
            self.kernel_process.kill()
            await self.kernel_process.wait()
        except Exception:
            pass
        try:
            os.remove(self.connection_file_path)
        except Exception:
            pass
        for task in self.channel_tasks:
            task.cancel()
        self.channel_tasks = []

    async def restart(self) -> None:
        await self.stop()
        self.setup_connection_file()
        await self.start()

    async def serve(
        self,
        websocket: AcceptedWebSocket,
        session_id: str,
        permissions: Optional[Dict[str, List[str]]],
    ):
        self.sessions[session_id] = websocket
        self.can_execute = permissions is None or "execute" in permissions.get("kernels", [])
        await self.listen_web(websocket)
        # the session could have been removed through the REST API, so check if it still exists
        if session_id in self.sessions:
            del self.sessions[session_id]

    async def listen_web(self, websocket: AcceptedWebSocket):
        try:
            await self.send_to_zmq(websocket)
        except WebSocketDisconnect:
            pass

    async def listen(self, channel_name: str):
        if channel_name == "shell":
            channel = self.shell_channel
        elif channel_name == "control":
            channel = self.control_channel
        elif channel_name == "iopub":
            channel = self.iopub_channel
        elif channel_name == "stdin":
            channel = self.stdin_channel

        while True:
            parts = await get_zmq_parts(channel)
            parent_header = get_parent_header(parts)
            if channel == self.iopub_channel:
                # broadcast to all web clients
                for websocket in self.sessions.values():
                    await self.send_to_ws(websocket, parts, parent_header, channel_name)
            else:
                session = parent_header["session"]
                if session in self.sessions:
                    websocket = self.sessions[session]
                    await self.send_to_ws(websocket, parts, parent_header, channel_name)

    async def _wait_for_ready(self):
        while True:
            msg = create_message("kernel_info_request")
            await send_message(msg, self.shell_channel, self.key)
            msg = await receive_message(self.shell_channel, timeout=0.2)
            if msg is not None and msg["msg_type"] == "kernel_info_reply":
                msg = await receive_message(self.iopub_channel, timeout=0.2)
                if msg is None:
                    # IOPub not connected, start over
                    pass
                else:
                    break

    async def send_to_zmq(self, websocket):
        if not websocket.accepted_subprotocol:
            while True:
                msg = await receive_json_or_bytes(websocket.websocket)
                if not self.can_execute:
                    continue
                msg_type = msg["header"]["msg_type"]
                if (msg_type in self.blocked_messages) or (
                    self.allowed_messages is not None and msg_type not in self.allowed_messages
                ):
                    continue
                channel = msg.pop("channel")
                if channel == "shell":
                    await send_message(msg, self.shell_channel, self.key)
                elif channel == "control":
                    await send_message(msg, self.control_channel, self.key)
                elif channel == "stdin":
                    await send_message(msg, self.stdin_channel, self.key)
        elif websocket.accepted_subprotocol == "v1.kernel.websocket.jupyter.org":
            while True:
                msg = await websocket.websocket.receive_bytes()
                if not self.can_execute:
                    continue
                channel, parts = deserialize_msg_from_ws_v1(msg)
                # NOTE: we parse the header for message filtering
                # it is not as bad as parsing the content
                header = json.loads(parts[0])
                msg_type = header["msg_type"]
                if (msg_type in self.blocked_messages) or (
                    self.allowed_messages is not None and msg_type not in self.allowed_messages
                ):
                    continue
                if channel == "shell":
                    await send_raw_message(parts, self.shell_channel, self.key)
                elif channel == "control":
                    await send_raw_message(parts, self.control_channel, self.key)
                elif channel == "stdin":
                    await send_raw_message(parts, self.stdin_channel, self.key)

    async def send_to_ws(self, websocket, parts, parent_header, channel_name):
        if not websocket.accepted_subprotocol:
            # default, "legacy" protocol
            msg = get_msg_from_parts(parts, parent_header=parent_header)
            msg["channel"] = channel_name
            await send_json_or_bytes(websocket.websocket, msg)
            if channel_name == "iopub":
                if "content" in msg and "execution_state" in msg["content"]:
                    self.last_activity = {
                        "date": msg["header"]["date"],
                        "execution_state": msg["content"]["execution_state"],
                    }
        elif websocket.accepted_subprotocol == "v1.kernel.websocket.jupyter.org":
            bin_msg = serialize_msg_to_ws_v1(parts, channel_name)
            try:
                await websocket.websocket.send_bytes(bin_msg)
            except Exception:
                pass
            # FIXME: update last_activity
            # but we don't want to parse the content!
            # or should we request it from the control channel?


async def receive_json_or_bytes(websocket):
    assert websocket.application_state == WebSocketState.CONNECTED
    message = await websocket.receive()
    websocket._raise_on_disconnect(message)
    if "text" in message:
        return json.loads(message["text"])
    msg = from_binary(message["bytes"])
    return msg


async def send_json_or_bytes(websocket, msg):
    bmsg = to_binary(msg)
    if bmsg is None:
        await websocket.send_json(msg)
    else:
        await websocket.send_bytes(bmsg)



import json
import os
from glob import glob
from pathlib import Path
from typing import List, Tuple


def get_federated_extensions(extensions_dir: Path) -> Tuple[List, List]:
    federated_extensions = []
    disabled_extensions = []

    for path in glob(os.path.join(extensions_dir, "**", "package.json"), recursive=True):
        with open(path) as f:
            package = json.load(f)
        if "jupyterlab" not in package:
            continue
        extension = package["jupyterlab"]["_build"]
        extension["name"] = package["name"]
        extension["description"] = package["description"]
        federated_extensions.append(extension)

        for ext in package["jupyterlab"].get("disabledExtensions", []):
            disabled_extensions.append(ext)

    return federated_extensions, disabled_extensions



import json
from typing import Dict, List

from fastapi import APIRouter, Depends, Query, Request, Response
from fastapi.responses import RedirectResponse
from fief_client import FiefAccessTokenInfo
from fps.hooks import register_router

from .backend import SESSION_COOKIE_NAME, auth, current_user, fief
from .models import UserRead

router = APIRouter()


@router.get("/auth-callback", name="auth_callback")
async def auth_callback(request: Request, response: Response, code: str = Query(...)):
    redirect_uri = request.url_for("auth_callback")
    tokens, _ = await fief.auth_callback(code, redirect_uri)

    response = RedirectResponse(request.url_for("root"))
    response.set_cookie(
        SESSION_COOKIE_NAME,
        tokens["access_token"],
        max_age=tokens["expires_in"],
        httponly=True,
        secure=False,
    )

    return response


@router.get("/api/me")
async def get_api_me(
    request: Request,
    user: UserRead = Depends(current_user()),
    access_token_info: FiefAccessTokenInfo = Depends(auth.authenticated()),
):
    checked_permissions: Dict[str, List[str]] = {}
    permissions = json.loads(dict(request.query_params).get("permissions", "{}").replace("'", '"'))
    if permissions:
        user_permissions = {}
        for permission in access_token_info["permissions"]:
            resource, action = permission.split(":")
            if resource not in user_permissions.keys():
                user_permissions[resource] = []
            user_permissions[resource].append(action)
        for resource, actions in permissions.items():
            user_resource_permissions = user_permissions.get(resource, [])
            allowed = checked_permissions[resource] = []
            for action in actions:
                if action in user_resource_permissions:
                    allowed.append(action)

    keys = ["username", "name", "display_name", "initials", "avatar_url", "color"]
    identity = {k: getattr(user, k) for k in keys}
    return {
        "identity": identity,
        "permissions": checked_permissions,
    }


r = register_router(router)



import json
import sys
from http import HTTPStatus
from pathlib import Path

import json5  # type: ignore
import pkg_resources  # type: ignore
from babel import Locale  # type: ignore
from fastapi import Depends, Response, status
from fastapi.responses import FileResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from fps_auth_base import User, current_user, update_user  # type: ignore
from fps_frontend.config import get_frontend_config  # type: ignore
from starlette.requests import Request  # type: ignore

import jupyverse  # type: ignore

from .utils import get_federated_extensions

try:
    import jupyterlab  # type: ignore
    from fps_jupyterlab.config import get_jlab_config  # type: ignore

    jlab_dev_mode = get_jlab_config().dev_mode
except Exception:
    jlab_dev_mode = False


LOCALE = "en"


prefix_dir = Path(sys.prefix)
if jlab_dev_mode:
    jlab_dir = Path(jupyterlab.__file__).parents[1] / "dev_mode"
else:
    jlab_dir = prefix_dir / "share" / "jupyter" / "lab"


def init_router(router, redirect_after_root):
    extensions_dir = prefix_dir / "share" / "jupyter" / "labextensions"
    federated_extensions, disabled_extensions = get_federated_extensions(extensions_dir)

    for ext in federated_extensions:
        name = ext["name"]
        router.mount(
            f"/lab/extensions/{name}/static",
            StaticFiles(directory=extensions_dir / name / "static"),
            name=name,
        )

    router.mount(
        "/lab/api/themes",
        StaticFiles(directory=jlab_dir / "themes"),
        name="themes",
    )

    @router.get("/", name="root")
    async def get_root(
        response: Response,
        frontend_config=Depends(get_frontend_config),
        user: User = Depends(current_user()),
    ):
        # auto redirect
        response.status_code = status.HTTP_302_FOUND
        response.headers["Location"] = frontend_config.base_url + redirect_after_root

    @router.get("/favicon.ico")
    async def get_favicon():
        return FileResponse(Path(jupyverse.__file__).parent / "static" / "favicon.ico")

    @router.get("/static/notebook/components/MathJax/{rest_of_path:path}")
    async def get_mathjax(rest_of_path):
        return RedirectResponse(
            "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.7/" + rest_of_path
        )

    @router.get("/lab/api/listings/@jupyterlab/extensionmanager-extension/listings.json")
    async def get_listings(user: User = Depends(current_user())):
        return {
            "blocked_extensions_uris": [],
            "allowed_extensions_uris": [],
            "blocked_extensions": [],
            "allowed_extensions": [],
        }

    @router.get("/lab/api/extensions")
    async def get_extensions(user: User = Depends(current_user())):
        return federated_extensions

    @router.get("/lab/api/translations/")
    async def get_translations_(
        frontend_config=Depends(get_frontend_config),
        user: User = Depends(current_user()),
    ):
        return RedirectResponse(f"{frontend_config.base_url}lab/api/translations")

    @router.get("/lab/api/translations")
    async def get_translations(user: User = Depends(current_user())):
        locale = Locale.parse("en")
        data = {
            "en": {
                "displayName": locale.get_display_name(LOCALE).capitalize(),
                "nativeName": locale.get_display_name().capitalize(),
            }
        }
        for ep in pkg_resources.iter_entry_points(group="jupyterlab.languagepack"):
            locale = Locale.parse(ep.name)
            data[ep.name] = {
                "displayName": locale.get_display_name(LOCALE).capitalize(),
                "nativeName": locale.get_display_name().capitalize(),
            }
        return {"data": data, "message": ""}

    @router.get("/lab/api/translations/{language}")
    async def get_translation(
        language,
        user: User = Depends(current_user()),
    ):
        global LOCALE
        if language == "en":
            LOCALE = language
            return {}
        for ep in pkg_resources.iter_entry_points(group="jupyterlab.languagepack"):
            if ep.name == language:
                break
        else:
            return {"data": {}, "message": f"Language pack '{language}' not installed!"}
        LOCALE = language
        package = ep.load()
        data = {}
        for path in (Path(package.__file__).parent / "locale" / language / "LC_MESSAGES").glob(
            "*.json"
        ):
            with open(path) as f:
                data.update({path.stem: json.load(f)})
        return {"data": data, "message": ""}

    @router.get("/lab/api/settings/{name0}/{name1}:{name2}")
    async def get_setting(
        name0,
        name1,
        name2,
        user: User = Depends(current_user()),
    ):
        with open(jlab_dir / "static" / "package.json") as f:
            package = json.load(f)
        if name0 in ["@jupyterlab", "@retrolab"]:
            schemas_parent = jlab_dir
        else:
            schemas_parent = extensions_dir / name0 / name1
        with open(schemas_parent / "schemas" / name0 / name1 / f"{name2}.json") as f:
            schema = json.load(f)
        key = f"{name1}:{name2}"
        setting = {
            "id": f"@jupyterlab/{key}",
            "schema": schema,
            "version": package["version"],
            "raw": "{}",
            "settings": {},
            "last_modified": None,
            "created": None,
        }
        if user:
            user_settings = json.loads(user.settings)
            if key in user_settings:
                setting.update(user_settings[key])
                setting["settings"] = json5.loads(user_settings[key]["raw"])
        return setting

    @router.put(
        "/lab/api/settings/@jupyterlab/{name0}:{name1}",
        status_code=204,
    )
    async def change_setting(
        request: Request,
        name0,
        name1,
        user: User = Depends(current_user()),
        user_update=Depends(update_user),
    ):
        settings = json.loads(user.settings)
        settings[f"{name0}:{name1}"] = await request.json()
        await user_update({"settings": json.dumps(settings)})
        return Response(status_code=HTTPStatus.NO_CONTENT.value)

    @router.get("/lab/api/settings")
    async def get_settings(user: User = Depends(current_user())):
        with open(jlab_dir / "static" / "package.json") as f:
            package = json.load(f)
        if user:
            user_settings = json.loads(user.settings)
        else:
            user_settings = {}
        settings = []
        for path in (jlab_dir / "schemas" / "@jupyterlab").glob("*/*.json"):
            with open(path) as f:
                schema = json.load(f)
            key = f"{path.parent.name}:{path.stem}"
            setting = {
                "id": f"@jupyterlab/{key}",
                "schema": schema,
                "version": package["version"],
                "raw": "{}",
                "settings": {},
                "warning": None,
                "last_modified": None,
                "created": None,
            }
            if key in user_settings:
                setting.update(user_settings[key])
                setting["settings"] = json5.loads(user_settings[key]["raw"])
            settings.append(setting)
        return {"settings": settings}

    return prefix_dir, federated_extensions



import contextlib
import json
from typing import Dict, List

from fastapi import APIRouter, Depends, Request
from fastapi_users.exceptions import UserAlreadyExists
from fps.config import get_config  # type: ignore
from fps.hooks import register_router  # type: ignore
from fps.logging import get_configured_logger  # type: ignore
from fps_uvicorn.cli import add_query_params  # type: ignore
from fps_uvicorn.config import UvicornConfig  # type: ignore
from sqlalchemy import select  # type: ignore

from .backends import (
    cookie_authentication,
    current_user,
    fapi_users,
    get_user_manager,
    github_authentication,
    github_cookie_authentication,
)
from .config import get_auth_config
from .db import (
    User,
    async_session_maker,
    create_db_and_tables,
    get_async_session,
    get_user_db,
    secret,
)
from .models import UserCreate, UserRead, UserUpdate

logger = get_configured_logger("auth")

auth_config = get_auth_config()
if auth_config.mode == "token":
    add_query_params({"token": auth_config.token})

router = APIRouter()


get_async_session_context = contextlib.asynccontextmanager(get_async_session)
get_user_db_context = contextlib.asynccontextmanager(get_user_db)
get_user_manager_context = contextlib.asynccontextmanager(get_user_manager)


@contextlib.asynccontextmanager
async def _get_user_manager():
    async with get_async_session_context() as session:
        async with get_user_db_context(session) as user_db:
            async with get_user_manager_context(user_db) as user_manager:
                yield user_manager


async def create_user(**kwargs):
    async with _get_user_manager() as user_manager:
        await user_manager.create(UserCreate(**kwargs))


async def update_user(user, **kwargs):
    async with _get_user_manager() as user_manager:
        await user_manager.update(UserUpdate(**kwargs), user)


async def get_user_by_email(user_email):
    async with _get_user_manager() as user_manager:
        return await user_manager.get_by_email(user_email)


@router.on_event("startup")
async def startup():
    await create_db_and_tables()

    if auth_config.test:
        try:
            await create_user(
                username="admin@jupyter.com",
                email="admin@jupyter.com",
                password="jupyverse",
                permissions={"admin": ["read", "write"]},
            )
        except UserAlreadyExists:
            pass

    try:
        await create_user(
            username=auth_config.token,
            email=auth_config.global_email,
            password="",
            permissions={},
        )
    except UserAlreadyExists:
        global_user = await get_user_by_email(auth_config.global_email)
        await update_user(
            global_user,
            username=auth_config.token,
            permissions={},
        )

    if auth_config.mode == "token":
        uvicorn_config = get_config(UvicornConfig)
        logger.info("")
        logger.info("To access the server, copy and paste this URL:")
        logger.info(
            f"http://{uvicorn_config.host}:{uvicorn_config.port}/?token={auth_config.token}"
        )
        logger.info("")


@router.get("/auth/users")
async def get_users(user: UserRead = Depends(current_user(permissions={"admin": ["read"]}))):
    async with async_session_maker() as session:
        statement = select(User)
        users = (await session.execute(statement)).unique().all()
    return [usr.User for usr in users if usr.User.is_active]


@router.get("/api/me")
async def get_api_me(
    request: Request,
    user: UserRead = Depends(current_user()),
):
    checked_permissions: Dict[str, List[str]] = {}
    permissions = json.loads(dict(request.query_params).get("permissions", "{}").replace("'", '"'))
    if permissions:
        user_permissions = user.permissions
        for resource, actions in permissions.items():
            user_resource_permissions = user_permissions.get(resource)
            if user_resource_permissions is None:
                continue
            allowed = checked_permissions[resource] = []
            for action in actions:
                if action in user_resource_permissions:
                    allowed.append(action)

    keys = ["username", "name", "display_name", "initials", "avatar_url", "color"]
    identity = {k: getattr(user, k) for k in keys}
    return {
        "identity": identity,
        "permissions": checked_permissions,
    }


# redefine GET /me because we want our current_user dependency
# it is first defined in users_router and so it wins over the one in fapi_users.get_users_router
users_router = APIRouter()


@users_router.get("/me")
async def get_me(user: UserRead = Depends(current_user(permissions={"admin": ["read"]}))):
    return user


users_router.include_router(fapi_users.get_users_router(UserRead, UserUpdate))

# Cookie based auth login and logout
r_cookie_auth = register_router(fapi_users.get_auth_router(cookie_authentication), prefix="/auth")
r_register = register_router(
    fapi_users.get_register_router(UserRead, UserCreate),
    prefix="/auth",
    dependencies=[Depends(current_user(permissions={"admin": ["write"]}))],
)
r_user = register_router(users_router, prefix="/auth/user")

# GitHub OAuth register router
r_github = register_router(
    fapi_users.get_oauth_router(github_authentication, github_cookie_authentication, secret),
    prefix="/auth/github",
)

r = register_router(router)



import json
import os
import shutil
from datetime import datetime
from http import HTTPStatus
from pathlib import Path
from typing import Dict, List, Optional, Union, cast

from anyio import open_file
from fastapi import APIRouter, Depends, HTTPException, Response
from fps.hooks import register_router  # type: ignore
from fps_auth_base import User, current_user  # type: ignore
from starlette.requests import Request  # type: ignore

from .models import Checkpoint, Content, CreateContent, RenameContent, SaveContent

router = APIRouter()


@router.post(
    "/api/contents/{path:path}/checkpoints",
    status_code=201,
)
async def create_checkpoint(
    path, user: User = Depends(current_user(permissions={"contents": ["write"]}))
):
    src_path = Path(path)
    dst_path = Path(".ipynb_checkpoints") / f"{src_path.stem}-checkpoint{src_path.suffix}"
    try:
        dst_path.parent.mkdir(exist_ok=True)
        shutil.copyfile(src_path, dst_path)
    except Exception:
        # FIXME: return error code?
        return []
    mtime = get_file_modification_time(dst_path)
    return Checkpoint(**{"id": "checkpoint", "last_modified": mtime})


@router.post(
    "/api/contents{path:path}",
    status_code=201,
)
async def create_content(
    path: Optional[str],
    request: Request,
    user: User = Depends(current_user(permissions={"contents": ["write"]})),
):
    create_content = CreateContent(**(await request.json()))
    content_path = Path(create_content.path)
    if create_content.type == "notebook":
        available_path = get_available_path(content_path / "Untitled.ipynb")
        async with await open_file(available_path, "w") as f:
            await f.write(
                json.dumps({"cells": [], "metadata": {}, "nbformat": 4, "nbformat_minor": 5})
            )
        src_path = available_path
        dst_path = Path(".ipynb_checkpoints") / f"{src_path.stem}-checkpoint{src_path.suffix}"
        try:
            dst_path.parent.mkdir(exist_ok=True)
            shutil.copyfile(src_path, dst_path)
        except Exception:
            # FIXME: return error code?
            pass
    elif create_content.type == "directory":
        name = "Untitled Folder"
        available_path = get_available_path(content_path / name, sep=" ")
        available_path.mkdir(parents=True, exist_ok=True)
    else:
        assert create_content.ext is not None
        available_path = get_available_path(content_path / ("untitled" + create_content.ext))
        open(available_path, "w").close()

    return await read_content(available_path, False)


@router.get("/api/contents")
async def get_root_content(
    content: int,
    user: User = Depends(current_user(permissions={"contents": ["read"]})),
):
    return await read_content("", bool(content))


@router.get("/api/contents/{path:path}/checkpoints")
async def get_checkpoint(
    path, user: User = Depends(current_user(permissions={"contents": ["read"]}))
):
    src_path = Path(path)
    dst_path = Path(".ipynb_checkpoints") / f"{src_path.stem}-checkpoint{src_path.suffix}"
    if not dst_path.exists():
        return []
    mtime = get_file_modification_time(dst_path)
    return [Checkpoint(**{"id": "checkpoint", "last_modified": mtime})]


@router.get("/api/contents/{path:path}")
async def get_content(
    path: str,
    content: int = 0,
    user: User = Depends(current_user(permissions={"contents": ["read"]})),
):
    return await read_content(path, bool(content))


@router.put("/api/contents/{path:path}")
async def save_content(
    path,
    request: Request,
    response: Response,
    user: User = Depends(current_user(permissions={"contents": ["write"]})),
):
    content = SaveContent(**(await request.json()))
    try:
        await write_content(content)
    except Exception:
        raise HTTPException(status_code=404, detail=f"Error saving {content.path}")
    return await read_content(content.path, False)


@router.delete(
    "/api/contents/{path:path}",
    status_code=204,
)
async def delete_content(
    path,
    user: User = Depends(current_user(permissions={"contents": ["write"]})),
):
    p = Path(path)
    if p.exists():
        if p.is_dir():
            shutil.rmtree(p)
        else:
            p.unlink()
    return Response(status_code=HTTPStatus.NO_CONTENT.value)


@router.patch("/api/contents/{path:path}")
async def rename_content(
    path,
    request: Request,
    user: User = Depends(current_user(permissions={"contents": ["write"]})),
):
    rename_content = RenameContent(**(await request.json()))
    Path(path).rename(rename_content.path)
    return await read_content(rename_content.path, False)


def get_file_modification_time(path: Path):
    if path.exists():
        return datetime.utcfromtimestamp(path.stat().st_mtime).isoformat() + "Z"


def get_file_creation_time(path: Path):
    if path.exists():
        return datetime.utcfromtimestamp(path.stat().st_ctime).isoformat() + "Z"


def get_file_size(path: Path) -> Optional[int]:
    if path.exists():
        return path.stat().st_size
    raise HTTPException(status_code=404, detail="Item not found")


def is_file_writable(path: Path) -> bool:
    if path.exists():
        if path.is_dir():
            # FIXME
            return True
        else:
            return os.access(path, os.W_OK)
    return False


async def read_content(path: Union[str, Path], get_content: bool, as_json: bool = False) -> Content:
    if isinstance(path, str):
        path = Path(path)
    content: Optional[Union[str, Dict, List[Dict]]] = None
    if get_content:
        if path.is_dir():
            content = [
                (await read_content(subpath, get_content=False)).dict()
                for subpath in path.iterdir()
                if not subpath.name.startswith(".")
            ]
        elif path.is_file() or path.is_symlink():
            try:
                async with await open_file(path) as f:
                    content = await f.read()
                if as_json:
                    content = json.loads(content)
            except Exception:
                raise HTTPException(status_code=404, detail="Item not found")
    format: Optional[str] = None
    if path.is_dir():
        size = None
        type = "directory"
        format = "json"
        mimetype = None
    elif path.is_file() or path.is_symlink():
        size = get_file_size(path)
        if path.suffix == ".ipynb":
            type = "notebook"
            format = None
            mimetype = None
            if content is not None:
                nb: dict
                if as_json:
                    content = cast(Dict, content)
                    nb = content
                else:
                    content = cast(str, content)
                    nb = json.loads(content)
                for cell in nb["cells"]:
                    if "metadata" not in cell:
                        cell["metadata"] = {}
                    cell["metadata"].update({"trusted": False})
                if not as_json:
                    content = json.dumps(nb)
        elif path.suffix == ".json":
            type = "json"
            format = "text"
            mimetype = "application/json"
        else:
            type = "file"
            format = None
            mimetype = "text/plain"
    else:
        raise HTTPException(status_code=404, detail="Item not found")

    return Content(
        **{
            "name": path.name,
            "path": path.as_posix(),
            "last_modified": get_file_modification_time(path),
            "created": get_file_creation_time(path),
            "content": content,
            "format": format,
            "mimetype": mimetype,
            "size": size,
            "writable": is_file_writable(path),
            "type": type,
        }
    )


async def write_content(content: Union[SaveContent, Dict]) -> None:
    if not isinstance(content, SaveContent):
        content = SaveContent(**content)
    async with await open_file(content.path, "w") as f:
        if content.format == "json":
            dict_content = cast(Dict, content.content)
            if content.type == "notebook":
                # see https://github.com/jupyterlab/jupyterlab/issues/11005
                if "metadata" in dict_content and "orig_nbformat" in dict_content["metadata"]:
                    del dict_content["metadata"]["orig_nbformat"]
            await f.write(json.dumps(dict_content, indent=2))
        else:
            content.content = cast(str, content.content)
            await f.write(content.content)


def get_available_path(path: Path, sep: str = ""):
    directory = path.parent
    name = Path(path.name)
    i = None
    while True:
        if i is None:
            i_str = ""
            i = 1
        else:
            i_str = str(i)
            i += 1
        if i_str:
            i_str = sep + i_str
        available_path = directory / (name.stem + i_str + name.suffix)
        if not available_path.exists():
            return available_path


r = register_router(router)



from uuid import uuid4

import pytest  # type: ignore

from .config import AuthConfig, get_auth_config


@pytest.fixture
def auth_mode():
    return "token"


@pytest.fixture
def auth_config(auth_mode):
    yield AuthConfig.parse_obj({"mode": auth_mode, "test": True})


@pytest.fixture
def config_override(app, auth_config):
    async def override_get_config():
        return auth_config

    app.dependency_overrides[get_auth_config] = override_get_config


@pytest.fixture()
def permissions():
    return {}


@pytest.fixture()
def authenticated_client(client, permissions):
    # create a new user
    username = uuid4().hex
    # if logged in, log out
    first_time = True
    while True:
        response = client.get("/api/me")
        if response.status_code == 403:
            break
        assert first_time
        response = client.post("/auth/logout")
        assert response.status_code == 200
        first_time = False

    # register user
    register_body = {
        "email": username + "@example.com",
        "password": username,
        "username": username,
        "permissions": permissions,
    }
    response = client.post("/auth/register", json=register_body)
    # check that we cannot register if not logged in
    assert response.status_code == 403
    # log in as admin
    login_body = {"username": "admin@jupyter.com", "password": "jupyverse"}
    response = client.post("/auth/login", data=login_body)
    assert response.status_code == 200
    # register user
    response = client.post("/auth/register", json=register_body)
    assert response.status_code == 201

    # log out
    response = client.post("/auth/logout")
    assert response.status_code == 200
    # check that we can't get our identity, since we're not logged in
    response = client.get("/api/me")
    assert response.status_code == 403

    # log in with registered user
    login_body = {"username": username + "@example.com", "password": username}
    response = client.post("/auth/login", data=login_body)
    assert response.status_code == 200
    # we should now have a cookie
    assert "fastapiusersauth" in client.cookies
    # check our identity, since we're logged in
    response = client.get("/api/me", params={"permissions": permissions})
    assert response.status_code == 200
    me = response.json()
    assert me["identity"]["username"] == username
    # check our permissions
    assert me["permissions"] == permissions
    yield client



import json
import sys

import pytest

test_theme = {"raw": '{// jupyverse test\n"theme": "JupyterLab Dark"}'}


@pytest.mark.parametrize("auth_mode", ("noauth",))
@pytest.mark.skipif(
    sys.platform.startswith(("linux", "darwin")) and sys.version_info < (3, 8),
    reason="pytest-asyncio issue",
)
def test_settings(client):
    # get previous theme
    response = client.get("/lab/api/settings/@jupyterlab/apputils-extension:themes")
    assert response.status_code == 200
    theme = {"raw": json.loads(response.content)["raw"]}
    # put new theme
    response = client.put(
        "/lab/api/settings/@jupyterlab/apputils-extension:themes",
        data=json.dumps(test_theme),
    )
    assert response.status_code == 204
    # get new theme
    response = client.get("/lab/api/settings/@jupyterlab/apputils-extension:themes")
    assert response.status_code == 200
    assert json.loads(response.content)["raw"] == test_theme["raw"]
    # put previous theme back
    response = client.put(
        "/lab/api/settings/@jupyterlab/apputils-extension:themes",
        data=json.dumps(theme),
    )
    assert response.status_code == 204



from typing import Any, Dict, List, Optional, Tuple

from fastapi import Depends, HTTPException, Request, Response, WebSocket, status
from fastapi.security import APIKeyCookie
from fief_client import FiefAccessTokenInfo, FiefAsync, FiefUserInfo
from fief_client.integrations.fastapi import FiefAuth

from .config import get_auth_fief_config
from .models import UserRead


class CustomFiefAuth(FiefAuth):
    client: FiefAsync

    async def get_unauthorized_response(self, request: Request, response: Response):
        redirect_uri = request.url_for("auth_callback")
        auth_url = await self.client.auth_url(redirect_uri, scope=["openid"])
        raise HTTPException(
            status_code=status.HTTP_307_TEMPORARY_REDIRECT,
            headers={"Location": auth_url},
        )


auth_fief_config = get_auth_fief_config()

fief = FiefAsync(
    auth_fief_config.base_url,
    auth_fief_config.client_id,
    auth_fief_config.client_secret.get_secret_value(),
)

SESSION_COOKIE_NAME = "fps_auth_fief_user_session"
scheme = APIKeyCookie(name=SESSION_COOKIE_NAME, auto_error=False)
auth = CustomFiefAuth(fief, scheme)


async def update_user(
    user: FiefUserInfo = Depends(auth.current_user()),
    access_token_info: FiefAccessTokenInfo = Depends(auth.authenticated()),
):
    async def _(data: Dict[str, Any]) -> FiefUserInfo:
        user = await fief.update_profile(access_token_info["access_token"], {"fields": data})
        return user

    return _


def websocket_auth(permissions: Optional[Dict[str, List[str]]] = None):
    async def _(
        websocket: WebSocket,
    ) -> Optional[Tuple[WebSocket, Optional[Dict[str, List[str]]]]]:
        accept_websocket = False
        checked_permissions: Optional[Dict[str, List[str]]] = None
        if SESSION_COOKIE_NAME in websocket._cookies:
            access_token = websocket._cookies[SESSION_COOKIE_NAME]
            if permissions is None:
                accept_websocket = True
            else:
                checked_permissions = {}
                for resource, actions in permissions.items():
                    allowed = checked_permissions[resource] = []
                    for action in actions:
                        try:
                            await fief.validate_access_token(
                                access_token, required_permissions=[f"{resource}:{action}"]
                            )
                        except BaseException:
                            pass
                        else:
                            allowed.append(action)
                            accept_websocket = True
        if accept_websocket:
            return websocket, checked_permissions
        else:
            await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
            return None

    return _


def current_user(permissions=None):
    if permissions is not None:
        permissions = [
            f"{resource}:{action}"
            for resource, actions in permissions.items()
            for action in actions
        ]

    async def _(user: FiefUserInfo = Depends(auth.current_user(permissions=permissions))):
        return UserRead(**user["fields"])

    return _



import json
from pathlib import Path

import retrolab  # type: ignore
from fastapi import APIRouter, Depends
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fps.hooks import register_router  # type: ignore
from fps_auth_base import User, current_user  # type: ignore
from fps_frontend.config import get_frontend_config  # type: ignore
from fps_lab.config import get_lab_config  # type: ignore
from fps_lab.routes import init_router  # type: ignore
from fps_lab.utils import get_federated_extensions  # type: ignore

router = APIRouter()
prefix_dir, federated_extensions = init_router(router, "retro/tree")
retrolab_dir = Path(retrolab.__file__).parent

router.mount(
    "/static/retro",
    StaticFiles(directory=retrolab_dir / "static"),
    name="static",
)

router.mount(
    "/lab/extensions/@retrolab/lab-extension/static",
    StaticFiles(directory=retrolab_dir / "labextension" / "static"),
    name="labextension/static",
)

for path in (retrolab_dir / "labextension" / "static").glob("remoteEntry.*.js"):
    load = f"static/{path.name}"
    break
retro_federated_extensions = [
    {
        "extension": "./extension",
        "load": load,
        "name": "@retrolab/lab-extension",
        "style": "./style",
    }
]


@router.get("/retro/tree", response_class=HTMLResponse)
async def get_tree(
    user: User = Depends(current_user()),
    frontend_config=Depends(get_frontend_config),
    lab_config=Depends(get_lab_config),
):
    return get_index("Tree", "tree", lab_config.collaborative, frontend_config.base_url)


@router.get("/retro/notebooks/{path:path}", response_class=HTMLResponse)
async def get_notebook(
    path,
    user: User = Depends(current_user()),
    frontend_config=Depends(get_frontend_config),
    lab_config=Depends(get_lab_config),
):
    return get_index(path, "notebooks", lab_config.collaborative, frontend_config.base_url)


@router.get("/retro/edit/{path:path}", response_class=HTMLResponse)
async def edit_file(
    path,
    user: User = Depends(current_user()),
    frontend_config=Depends(get_frontend_config),
    lab_config=Depends(get_lab_config),
):
    return get_index(path, "edit", lab_config.collaborative, frontend_config.base_url)


@router.get("/retro/consoles/{path:path}", response_class=HTMLResponse)
async def get_console(
    path,
    user: User = Depends(current_user()),
    frontend_config=Depends(get_frontend_config),
    lab_config=Depends(get_lab_config),
):
    return get_index(path, "consoles", lab_config.collaborative, frontend_config.base_url)


@router.get("/retro/terminals/{name}", response_class=HTMLResponse)
async def get_terminal(
    name: str,
    user: User = Depends(current_user()),
    frontend_config=Depends(get_frontend_config),
    lab_config=Depends(get_lab_config),
):
    return get_index(name, "terminals", lab_config.collaborative, frontend_config.base_url)


def get_index(doc_name, retro_page, collaborative, base_url="/"):
    extensions_dir = prefix_dir / "share" / "jupyter" / "labextensions"
    federated_extensions, disabled_extension = get_federated_extensions(extensions_dir)
    page_config = {
        "appName": "RetroLab",
        "appNamespace": "retro",
        "appSettingsDir": (prefix_dir / "share" / "jupyter" / "lab" / "settings").as_posix(),
        "appUrl": "/lab",
        "appVersion": retrolab.__version__,
        "baseUrl": base_url,
        "cacheFiles": True,
        "collaborative": collaborative,
        "disabledExtensions": disabled_extension,
        "extraLabextensionsPath": [],
        "federated_extensions": retro_federated_extensions + federated_extensions,
        "frontendUrl": "/retro/",
        "fullAppUrl": f"{base_url}lab",
        "fullLabextensionsUrl": f"{base_url}lab/extensions",
        "fullLicensesUrl": f"{base_url}lab/api/licenses",
        "fullListingsUrl": f"{base_url}lab/api/listings",
        "fullMathjaxUrl": f"{base_url}static/notebook/components/MathJax/MathJax.js",
        "fullSettingsUrl": f"{base_url}lab/api/settings",
        "fullStaticUrl": f"{base_url}static/retro",
        "fullThemesUrl": f"{base_url}lab/api/themes",
        "fullTranslationsApiUrl": f"{base_url}lab/api/translations",
        "fullTreeUrl": f"{base_url}lab/tree",
        "fullWorkspacesApiUrl": f"{base_url}lab/api/workspaces",
        "labextensionsPath": [(prefix_dir / "share" / "jupyter" / "labextensions").as_posix()],
        "labextensionsUrl": "/lab/extensions",
        "licensesUrl": "/lab/api/licenses",
        "listingsUrl": "/lab/api/listings",
        "mathjaxConfig": "TeX-AMS-MML_HTMLorMML-full,Safe",
        "retroLogo": False,
        "retroPage": retro_page,
        "schemasDir": (prefix_dir / "share" / "jupyter" / "lab" / "schemas").as_posix(),
        "settingsUrl": "/lab/api/settings",
        "staticDir": (retrolab_dir / "static").as_posix(),
        "templatesDir": (retrolab_dir / "templates").as_posix(),
        "terminalsAvailable": True,
        "themesDir": (prefix_dir / "share" / "jupyter" / "lab" / "themes").as_posix(),
        "themesUrl": "/lab/api/themes",
        "translationsApiUrl": "/lab/api/translations",
        "treeUrl": "/lab/tree",
        "workspacesApiUrl": "/lab/api/workspaces",
        "wsUrl": "",
    }
    index = (
        INDEX_HTML.replace("PAGE_CONFIG", json.dumps(page_config))
        .replace("DOC_NAME", doc_name)
        .replace("BASE_URL", base_url)
    )
    return index


INDEX_HTML = """\
<!DOCTYPE html>
<html>
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <title>RetroLab - DOC_NAME</title>
  <link rel="icon" type="image/x-icon" href="/static/favicons/favicon-notebook.ico"
    class="favicon">
</head>
<body>
  <script id="jupyter-config-data" type="application/json">
    PAGE_CONFIG
  </script>
  <script src="BASE_URLstatic/retro/bundle.js" main="index"></script>
  <script type="text/javascript">
    /* Remove token from URL. */
    (function () {
      var parsedUrl = new URL(window.location.href);
      if (parsedUrl.searchParams.get('token')) {
        parsedUrl.searchParams.delete('token');
        window.history.replaceState({ }, '', parsedUrl.href);
      }
    })();
  </script>
</body>
</html>
"""

r = register_router(router)



import pytest
from fps_auth.config import get_auth_config
from starlette.websockets import WebSocketDisconnect


def test_kernel_channels_unauthenticated(client):
    with pytest.raises(WebSocketDisconnect):
        with client.websocket_connect(
            "/api/kernels/kernel_id_0/channels?session_id=session_id_0",
        ):
            pass


def test_kernel_channels_authenticated(authenticated_client):
    with authenticated_client.websocket_connect(
        "/api/kernels/kernel_id_0/channels?session_id=session_id_0",
    ):
        pass


@pytest.mark.parametrize("auth_mode", ("noauth", "token", "user"))
def test_root_auth(auth_mode, client):
    response = client.get("/")
    if auth_mode == "noauth":
        expected = 200
        content_type = "text/html; charset=utf-8"
    elif auth_mode in ["token", "user"]:
        expected = 403
        content_type = "application/json"

    assert response.status_code == expected
    assert response.headers["content-type"] == content_type


@pytest.mark.parametrize("auth_mode", ("noauth",))
def test_no_auth(client):
    response = client.get("/lab/api/settings")
    assert response.status_code == 200


@pytest.mark.parametrize("auth_mode", ("token",))
def test_token_auth(client):
    # no token provided, should not work
    response = client.get("/")
    assert response.status_code == 403
    # token provided, should work
    auth_config = get_auth_config()
    response = client.get(f"/?token={auth_config.token}")
    assert response.status_code == 200


@pytest.mark.parametrize("auth_mode", ("user",))
@pytest.mark.parametrize(
    "permissions",
    (
        {},
        {"admin": ["read"], "foo": ["bar", "baz"]},
    ),
)
def test_permissions(authenticated_client, permissions):
    response = authenticated_client.get("/auth/user/me")
    if "admin" in permissions.keys():
        # we have the permissions
        assert response.status_code == 200
    else:
        # we don't have the permissions
        assert response.status_code == 403



from fps.config import PluginModel, get_config  # type: ignore
from fps.hooks import register_config  # type: ignore
from pydantic import BaseSettings  # type: ignore


class ResourceUsageConfig(PluginModel, BaseSettings):
    mem_limit: int = 0
    mem_warning_threshold: int = 0
    track_cpu_percent: bool = False
    cpu_limit: int = 0
    cpu_warning_threshold: int = 0


def get_resource_usage_config():
    return get_config(ResourceUsageConfig)


c = register_config(ResourceUsageConfig)



from fps.config import PluginModel, get_config  # type: ignore
from fps.hooks import register_config  # type: ignore


class JupyterLabConfig(PluginModel):
    dev_mode: bool = False


def get_jlab_config():
    return get_config(JupyterLabConfig)


c = register_config(JupyterLabConfig)



import pkg_resources

__version__ = "0.0.48"

auth = {ep.name: ep.load() for ep in pkg_resources.iter_entry_points(group="jupyverse_auth")}

try:
    User = auth["User"]
    current_user = auth["current_user"]
    update_user = auth["update_user"]
    websocket_auth = auth["websocket_auth"]
except KeyError:
    raise RuntimeError(
        "An auth plugin must be installed, for instance: pip install fps-auth",
    )



import os
from datetime import datetime
from http import HTTPStatus
from typing import Any, Dict

from fastapi import APIRouter, Depends, Response
from fps.hooks import register_router  # type: ignore
from fps_auth_base import User, current_user, websocket_auth  # type: ignore

from .models import Terminal

if os.name == "nt":
    from .win_server import TerminalServer  # type: ignore
else:
    from .server import TerminalServer  # type: ignore

router = APIRouter()

TERMINALS: Dict[str, Dict[str, Any]] = {}


@router.get("/api/terminals")
async def get_terminals(user: User = Depends(current_user({"terminals": ["read"]}))):
    return [terminal["info"] for terminal in TERMINALS.values()]


@router.post("/api/terminals")
async def create_terminal(
    user: User = Depends(current_user({"terminals": ["write"]})),
):
    name = str(len(TERMINALS) + 1)
    terminal = Terminal(
        **{
            "name": name,
            "last_activity": datetime.utcnow().isoformat() + "Z",
        }
    )
    server = TerminalServer()
    TERMINALS[name] = {"info": terminal, "server": server}
    return terminal


@router.delete("/api/terminals/{name}", status_code=204)
async def delete_terminal(
    name: str,
    user: User = Depends(current_user(permissions={"terminals": ["write"]})),
):
    for websocket in TERMINALS[name]["server"].websockets:
        TERMINALS[name]["server"].quit(websocket)
    del TERMINALS[name]
    return Response(status_code=HTTPStatus.NO_CONTENT.value)


@router.websocket("/terminals/websocket/{name}")
async def terminal_websocket(
    name,
    websocket_permissions=Depends(websocket_auth(permissions={"terminals": ["read", "execute"]})),
):
    if websocket_permissions is None:
        return
    websocket, permissions = websocket_permissions
    await websocket.accept()
    await TERMINALS[name]["server"].serve(websocket, permissions)
    if name in TERMINALS:
        TERMINALS[name]["server"].quit(websocket)
        if not TERMINALS[name]["server"].websockets:
            del TERMINALS[name]


r = register_router(router)



import os
from pathlib import Path

import pytest
from utils import clear_content_values, create_content, sort_content_by_name


@pytest.mark.parametrize("auth_mode", ("noauth",))
def test_tree(client, tmp_path):
    prev_dir = os.getcwd()
    os.chdir(tmp_path)
    dname = Path(".")
    expected = []
    # write some files with different size
    for size in range(7, 13):
        fname = f"file_size_{size}"
        with (dname / fname).open("w") as f:
            f.write("-" * size)
        expected.append(
            create_content(
                content=None,
                type="file",
                size=size,
                mimetype="text/plain",
                name=fname,
                path=(dname / fname).as_posix(),
                format=None,
            )
        )
    # write some directories
    for i in range(3):
        sub_dname = f"directory_{i}"
        (dname / sub_dname).mkdir()
        expected.append(
            create_content(
                content=None,
                type="directory",
                size=None,
                mimetype=None,
                name=sub_dname,
                path=(dname / sub_dname).as_posix(),
                format="json",
            )
        )
    expected = create_content(
        content=expected,
        type="directory",
        size=None,
        mimetype=None,
        name="",
        path=dname.as_posix(),
        format="json",
    )
    response = client.get("/api/contents", params={"content": 1})
    actual = response.json()
    # ignore modification and creation times
    clear_content_values(actual, keys=["created", "last_modified"])
    # ensure content names are ordered the same way
    sort_content_by_name(actual)
    sort_content_by_name(expected)
    assert actual == expected
    os.chdir(prev_dir)



from fps.config import PluginModel, get_config  # type: ignore
from fps.hooks import register_config  # type: ignore


class LabConfig(PluginModel):
    collaborative: bool = False


def get_lab_config():
    return get_config(LabConfig)


c = register_config(LabConfig)



import sys
from pathlib import Path
from time import sleep

import pytest
from fps_kernels.kernel_server.server import KernelServer, kernels


@pytest.mark.asyncio
@pytest.mark.parametrize("auth_mode", ("noauth",))
@pytest.mark.skipif(
    sys.version_info < (3, 8) or sys.platform.startswith("win"),
    reason="pytest-asyncio issue",
)
async def test_kernel_messages(client, capfd):
    kernel_id = "kernel_id_0"
    kernel_name = "python3"
    kernelspec_path = (
        Path(sys.prefix) / "share" / "jupyter" / "kernels" / kernel_name / "kernel.json"
    )
    assert kernelspec_path.exists()
    kernel_server = KernelServer(kernelspec_path=kernelspec_path, capture_kernel_output=False)
    await kernel_server.start()
    kernels[kernel_id] = {"server": kernel_server}
    msg_id = "0"
    msg = {
        "channel": "shell",
        "parent_header": None,
        "content": None,
        "metadata": None,
        "header": {
            "msg_type": "msg_type_0",
            "msg_id": msg_id,
        },
    }

    # block msg_type_0
    msg["header"]["msg_id"] = str(int(msg["header"]["msg_id"]) + 1)
    kernel_server.block_messages("msg_type_0")
    with client.websocket_connect(
        f"/api/kernels/{kernel_id}/channels?session_id=session_id_0",
    ) as websocket:
        websocket.send_json(msg)
    sleep(0.5)
    out, err = capfd.readouterr()
    assert not err

    # allow only msg_type_0
    msg["header"]["msg_id"] = str(int(msg["header"]["msg_id"]) + 1)
    kernel_server.allow_messages("msg_type_0")
    with client.websocket_connect(
        f"/api/kernels/{kernel_id}/channels?session_id=session_id_0",
    ) as websocket:
        websocket.send_json(msg)
    sleep(0.5)
    out, err = capfd.readouterr()
    assert err.count("[IPKernelApp] WARNING | Unknown message type: 'msg_type_0'") == 1

    # block all messages
    msg["header"]["msg_id"] = str(int(msg["header"]["msg_id"]) + 1)
    kernel_server.allow_messages([])
    with client.websocket_connect(
        f"/api/kernels/{kernel_id}/channels?session_id=session_id_0",
    ) as websocket:
        websocket.send_json(msg)
    sleep(0.5)
    out, err = capfd.readouterr()
    assert not err

    # allow all messages
    msg["header"]["msg_id"] = str(int(msg["header"]["msg_id"]) + 1)
    kernel_server.allow_messages()
    with client.websocket_connect(
        f"/api/kernels/{kernel_id}/channels?session_id=session_id_0",
    ) as websocket:
        websocket.send_json(msg)
    sleep(0.5)
    out, err = capfd.readouterr()
    assert err.count("[IPKernelApp] WARNING | Unknown message type: 'msg_type_0'") == 1



import psutil  # type: ignore
from anyio import to_thread
from fastapi import APIRouter, Depends  # type: ignore
from fps.hooks import register_router  # type: ignore
from fps_auth_base import User, current_user  # type: ignore

from .config import get_resource_usage_config

router = APIRouter()


@router.get("/api/metrics/v1")
async def get_content(
    user: User = Depends(current_user(permissions={"contents": ["read"]})),
    config=Depends(get_resource_usage_config),
):
    cur_process = psutil.Process()
    all_processes = [cur_process] + cur_process.children(recursive=True)

    # Get memory information
    rss = 0
    for p in all_processes:
        try:
            rss += p.memory_info().rss
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            pass

    mem_limit = config.mem_limit

    limits = {"memory": {"rss": mem_limit}}
    if config.mem_limit and config.mem_warning_threshold:
        limits["memory"]["warn"] = (mem_limit - rss) < (mem_limit * config.mem_warning_threshold)

    metrics = {"rss": rss, "limits": limits}

    # Optionally get CPU information
    if config.track_cpu_percent:
        cpu_count = psutil.cpu_count()
        cpu_percent = await to_thread.run_sync(_get_cpu_percent, all_processes)

        if config.cpu_limit:
            limits["cpu"] = {"cpu": config.cpu_limit}
            if config.cpu_warning_threshold:
                limits["cpu"]["warn"] = (config.cpu_limit - cpu_percent) < (
                    config.cpu_limit * config.cpu_warning_threshold
                )

        metrics.update(cpu_percent=cpu_percent, cpu_count=cpu_count)

    return metrics


def _get_cpu_percent(all_processes):
    def get_cpu_percent(p):
        try:
            return p.cpu_percent(interval=0.05)
        # Avoid littering logs with stack traces complaining
        # about dead processes having no CPU usage
        except BaseException:
            return 0

    return sum([get_cpu_percent(p) for p in all_processes])


r = register_router(router)



import json
from http import HTTPStatus
from pathlib import Path

import jupyterlab  # type: ignore
from fastapi import APIRouter, Depends, Response
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fps.hooks import register_router  # type: ignore
from fps_auth_base import User, current_user, update_user  # type: ignore
from fps_frontend.config import get_frontend_config  # type: ignore
from fps_lab.config import get_lab_config  # type: ignore
from fps_lab.routes import init_router  # type: ignore
from fps_lab.utils import get_federated_extensions  # type: ignore
from starlette.requests import Request  # type: ignore

from .config import get_jlab_config

router = APIRouter()
prefix_dir, federated_extensions = init_router(router, "lab")
jupyterlab_dir = Path(jupyterlab.__file__).parents[1]

config = get_jlab_config()
if config.dev_mode:
    static_lab_dir = jupyterlab_dir / "dev_mode" / "static"
else:
    static_lab_dir = prefix_dir / "share" / "jupyter" / "lab" / "static"

router.mount(
    "/static/lab",
    StaticFiles(directory=static_lab_dir),
    name="static",
)


@router.get("/lab")
async def get_lab(
    user: User = Depends(current_user()),
    frontend_config=Depends(get_frontend_config),
    lab_config=Depends(get_lab_config),
):
    return HTMLResponse(
        get_index(
            "default",
            lab_config.collaborative,
            config.dev_mode,
            frontend_config.base_url,
        )
    )


@router.get("/lab/tree/{path:path}")
async def load_workspace(
    path,
    frontend_config=Depends(get_frontend_config),
    lab_config=Depends(get_lab_config),
):
    return HTMLResponse(
        get_index(
            "default",
            lab_config.collaborative,
            config.dev_mode,
            frontend_config.base_url,
        )
    )


@router.get("/lab/api/workspaces/{name}")
async def get_workspace_data(user: User = Depends(current_user())):
    if user:
        return json.loads(user.workspace)
    return {}


@router.put(
    "/lab/api/workspaces/{name}",
    status_code=204,
)
async def set_workspace(
    request: Request,
    user: User = Depends(current_user()),
    user_update=Depends(update_user),
):
    workspace = (await request.body()).decode("utf-8")
    await user_update({"workspace": workspace})
    return Response(status_code=HTTPStatus.NO_CONTENT.value)


@router.get("/lab/workspaces/{name}", response_class=HTMLResponse)
async def get_workspace(
    name,
    user: User = Depends(current_user()),
    frontend_config=Depends(get_frontend_config),
    lab_config=Depends(get_lab_config),
):
    return get_index(name, lab_config.collaborative, config.dev_mode, frontend_config.base_url)


INDEX_HTML = """\
<!doctype html><html lang="en"><head><meta charset="utf-8"><title>JupyterLab</title>
<meta name="viewport" content="width=device-width,initial-scale=1">
<script id="jupyter-config-data" type="application/json">PAGE_CONFIG</script>
VENDORS_NODE_MODULES
<script defer="defer" src="FULL_STATIC_URL/main.MAIN_ID.js?v=MAIN_ID"></script>
</head><body><script>/* Remove token from URL. */
  (function () {
    var location = window.location;
    var search = location.search;

    // If there is no query string, bail.
    if (search.length <= 1) {
      return;
    }

    // Rebuild the query string without the `token`.
    var query = '?' + search.slice(1).split('&')
      .filter(function (param) { return param.split('=')[0] !== 'token'; })
      .join('&');

    // Rebuild the URL with the new query string.
    var url = location.origin + location.pathname +
      (query !== '?' ? query : '') + location.hash;

    if (url === location.href) {
      return;
    }

    window.history.replaceState({ }, '', url);
  })();</script></body></html>
"""


def get_index(workspace, collaborative, dev_mode, base_url="/"):
    for path in (static_lab_dir).glob("main.*.js"):
        main_id = path.name.split(".")[1]
        break
    vendor_id = None
    for path in (static_lab_dir).glob("vendors-node_modules_whatwg-fetch_fetch_js.*.js"):
        vendor_id = path.name.split(".")[1]
        break
    full_static_url = f"{base_url}static/lab"
    extensions_dir = prefix_dir / "share" / "jupyter" / "labextensions"
    federated_extensions, disabled_extension = get_federated_extensions(extensions_dir)

    page_config = {
        "appName": "JupyterLab",
        "appNamespace": "lab",
        "appUrl": "/lab",
        "appVersion": jupyterlab.__version__,
        "baseUrl": base_url,
        "cacheFiles": False,
        "collaborative": collaborative,
        "devMode": dev_mode,
        "disabledExtensions": disabled_extension,
        "exposeAppInBrowser": False,
        "extraLabextensionsPath": [],
        "federated_extensions": federated_extensions,
        "fullAppUrl": f"{base_url}lab",
        "fullLabextensionsUrl": f"{base_url}lab/extensions",
        "fullLicensesUrl": f"{base_url}lab/api/licenses",
        "fullListingsUrl": f"{base_url}lab/api/listings",
        "fullMathjaxUrl": f"{base_url}static/notebook/components/MathJax/MathJax.js",
        "fullSettingsUrl": f"{base_url}lab/api/settings",
        "fullStaticUrl": full_static_url,
        "fullThemesUrl": f"{base_url}lab/api/themes",
        "fullTranslationsApiUrl": f"{base_url}lab/api/translations",
        "fullTreeUrl": f"{base_url}lab/tree",
        "fullWorkspacesApiUrl": f"{base_url}lab/api/workspaces",
        "ignorePlugins": [],
        "labextensionsUrl": "/lab/extensions",
        "licensesUrl": "/lab/api/licenses",
        "listingsUrl": "/lab/api/listings",
        "mathjaxConfig": "TeX-AMS-MML_HTMLorMML-full,Safe",
        "mode": "multiple-document",
        "notebookVersion": "[1, 9, 0]",
        "quitButton": True,
        "settingsUrl": "/lab/api/settings",
        "store_id": 0,
        "schemasDir": (prefix_dir / "share" / "jupyter" / "lab" / "schemas").as_posix(),
        "terminalsAvailable": True,
        "themesDir": (prefix_dir / "share" / "jupyter" / "lab" / "themes").as_posix(),
        "themesUrl": "/lab/api/themes",
        "token": "4e2804532de366abc81e32ab0c6bf68a73716fafbdbb2098",
        "translationsApiUrl": "/lab/api/translations",
        "treePath": "",
        "workspace": workspace,
        "treeUrl": "/lab/tree",
        "workspacesApiUrl": "/lab/api/workspaces",
        "wsUrl": "",
    }
    index = (
        INDEX_HTML.replace("PAGE_CONFIG", json.dumps(page_config))
        .replace("FULL_STATIC_URL", full_static_url)
        .replace("MAIN_ID", main_id)
    )
    if vendor_id:
        index = index.replace(
            "VENDORS_NODE_MODULES",
            '<script defer src="/static/lab/vendors-node_modules_whatwg-fetch_fetch_js.'
            f'{vendor_id}.js"></script>',
        )
    else:
        index = index.replace("VENDORS_NODE_MODULES", "")
    return index


r = register_router(router)



from fps.config import PluginModel, get_config  # type: ignore
from fps.hooks import register_config  # type: ignore


class RetroLabConfig(PluginModel):
    pass


def get_rlab_config():
    return get_config(RetroLabConfig)


c = register_config(RetroLabConfig)



import uuid
from typing import Any, Dict, Generic, List, Optional, Tuple

import httpx
from fastapi import Depends, HTTPException, Response, WebSocket, status
from fastapi_users import (  # type: ignore
    BaseUserManager,
    FastAPIUsers,
    UUIDIDMixin,
    models,
)
from fastapi_users.authentication import (
    AuthenticationBackend,
    CookieTransport,
    JWTStrategy,
)
from fastapi_users.authentication.strategy.base import Strategy
from fastapi_users.authentication.transport.base import Transport
from fastapi_users.db import SQLAlchemyUserDatabase
from fps.exceptions import RedirectException  # type: ignore
from fps.logging import get_configured_logger  # type: ignore
from fps_lab.config import get_lab_config  # type: ignore
from httpx_oauth.clients.github import GitHubOAuth2  # type: ignore
from starlette.requests import Request

from .config import get_auth_config
from .db import User, get_user_db, secret
from .models import UserCreate, UserRead

logger = get_configured_logger("auth")


class NoAuthTransport(Transport):
    scheme = None  # type: ignore

    async def get_login_response(self, token: str, response: Response):
        pass

    async def get_logout_response(self, response: Response):
        pass

    @staticmethod
    def get_openapi_login_responses_success():
        pass

    @staticmethod
    def get_openapi_logout_responses_success():
        pass


class NoAuthStrategy(Strategy, Generic[models.UP, models.ID]):
    async def read_token(
        self, token: Optional[str], user_manager: BaseUserManager[models.UP, models.ID]
    ) -> Optional[models.UP]:
        active_user = await user_manager.user_db.get_by_email(get_auth_config().global_email)
        return active_user

    async def write_token(self, user: models.UP):
        pass

    async def destroy_token(self, token: str, user: models.UP):
        pass


class GitHubTransport(CookieTransport):
    async def get_login_response(self, token: str, response: Response):
        await super().get_login_response(token, response)
        response.status_code = status.HTTP_302_FOUND
        response.headers["Location"] = "/lab"


noauth_transport = NoAuthTransport()
cookie_transport = CookieTransport(cookie_secure=get_auth_config().cookie_secure)
github_transport = GitHubTransport()


def get_noauth_strategy() -> NoAuthStrategy:
    return NoAuthStrategy()


def get_jwt_strategy() -> JWTStrategy:
    return JWTStrategy(secret=secret, lifetime_seconds=None)  # type: ignore


noauth_authentication = AuthenticationBackend(
    name="noauth",
    transport=noauth_transport,
    get_strategy=get_noauth_strategy,
)
cookie_authentication = AuthenticationBackend(
    name="cookie",
    transport=cookie_transport,
    get_strategy=get_jwt_strategy,
)
github_cookie_authentication = AuthenticationBackend(
    name="github",
    transport=github_transport,
    get_strategy=get_jwt_strategy,
)
github_authentication = GitHubOAuth2(
    get_auth_config().client_id, get_auth_config().client_secret.get_secret_value()
)


class UserManager(UUIDIDMixin, BaseUserManager[User, uuid.UUID]):
    async def on_after_register(self, user: User, request: Optional[Request] = None):
        for oauth_account in user.oauth_accounts:
            if oauth_account.oauth_name == "github":
                async with httpx.AsyncClient() as client:
                    r = (
                        await client.get(f"https://api.github.com/user/{oauth_account.account_id}")
                    ).json()

                await self.user_db.update(
                    user,
                    dict(
                        anonymous=False,
                        username=r["login"],
                        color=None,
                        avatar_url=r["avatar_url"],
                        is_active=True,
                    ),
                )


async def get_user_manager(user_db: SQLAlchemyUserDatabase = Depends(get_user_db)):
    yield UserManager(user_db)


async def get_enabled_backends(
    auth_config=Depends(get_auth_config), lab_config=Depends(get_lab_config)
):
    if auth_config.mode == "noauth" and not lab_config.collaborative:
        return [noauth_authentication, github_cookie_authentication]
    else:
        return [cookie_authentication, github_cookie_authentication]


fapi_users = FastAPIUsers[User, uuid.UUID](
    get_user_manager,
    [noauth_authentication, cookie_authentication, github_cookie_authentication],
)


async def create_guest(user_manager, auth_config):
    # workspace and settings are copied from global user
    # but this is a new user
    global_user = await user_manager.get_by_email(auth_config.global_email)
    user_id = str(uuid.uuid4())
    guest = dict(
        anonymous=True,
        email=f"{user_id}@jupyter.com",
        username=f"{user_id}@jupyter.com",
        password="",
        workspace=global_user.workspace,
        settings=global_user.settings,
        permissions={},
    )
    return await user_manager.create(UserCreate(**guest))


def current_user(permissions: Optional[Dict[str, List[str]]] = None):
    async def _(
        response: Response,
        token: Optional[str] = None,
        user: Optional[User] = Depends(
            fapi_users.current_user(optional=True, get_enabled_backends=get_enabled_backends)
        ),
        user_manager: UserManager = Depends(get_user_manager),
        auth_config=Depends(get_auth_config),
        lab_config=Depends(get_lab_config),
    ):
        if auth_config.mode == "user":
            # "user" authentication: check authorization
            if user and permissions:
                for resource, actions in permissions.items():
                    user_actions_for_resource = user.permissions.get(resource, [])
                    if not all([a in user_actions_for_resource for a in actions]):
                        user = None
                        break
        else:
            # "noauth" or "token" authentication
            if lab_config.collaborative:
                if not user and auth_config.mode == "noauth":
                    user = await create_guest(user_manager, auth_config)
                    await cookie_authentication.login(get_jwt_strategy(), user, response)

                elif not user and auth_config.mode == "token":
                    global_user = await user_manager.get_by_email(auth_config.global_email)
                    if global_user and global_user.username == token:
                        user = await create_guest(user_manager, auth_config)
                        await cookie_authentication.login(get_jwt_strategy(), user, response)
            else:
                if auth_config.mode == "token":
                    global_user = await user_manager.get_by_email(auth_config.global_email)
                    if global_user and global_user.username == token:
                        user = global_user
                        await cookie_authentication.login(get_jwt_strategy(), user, response)

        if user:
            return user

        elif auth_config.login_url:
            raise RedirectException(auth_config.login_url)

        else:
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)

    return _


def websocket_auth(permissions: Optional[Dict[str, List[str]]] = None):
    """
    A function returning a dependency for the WebSocket connection.

    :param permissions: the permissions the user should be granted access to. The user should have
    access to at least one of them for the WebSocket to be opened.
    :returns: a dependency for the WebSocket connection. The dependency returns a tuple consisting
    of the websocket and the checked user permissions if the websocket is accepted, None otherwise.
    """

    async def _(
        websocket: WebSocket,
        auth_config=Depends(get_auth_config),
        user_manager: UserManager = Depends(get_user_manager),
    ) -> Optional[Tuple[WebSocket, Optional[Dict[str, List[str]]]]]:
        accept_websocket = False
        checked_permissions: Optional[Dict[str, List[str]]] = None
        if auth_config.mode == "noauth":
            accept_websocket = True
        elif "fastapiusersauth" in websocket._cookies:
            token = websocket._cookies["fastapiusersauth"]
            user = await get_jwt_strategy().read_token(token, user_manager)
            if user:
                if auth_config.mode == "user":
                    # "user" authentication: check authorization
                    if permissions is None:
                        accept_websocket = True
                    else:
                        checked_permissions = {}
                        for resource, actions in permissions.items():
                            user_actions_for_resource = user.permissions.get(resource)
                            if user_actions_for_resource is None:
                                continue
                            allowed = checked_permissions[resource] = []
                            for action in actions:
                                if action in user_actions_for_resource:
                                    allowed.append(action)
                                    accept_websocket = True
                else:
                    accept_websocket = True
        if accept_websocket:
            return websocket, checked_permissions
        else:
            await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
            return None

    return _


async def update_user(
    user: UserRead = Depends(current_user()), user_db: SQLAlchemyUserDatabase = Depends(get_user_db)
):
    async def _(data: Dict[str, Any]) -> UserRead:
        await user_db.update(user, data)
        return user

    return _



from fps.config import PluginModel, get_config
from fps.hooks import register_config
from pydantic import BaseSettings, SecretStr


class AuthFiefConfig(PluginModel, BaseSettings):
    base_url: str  # Base URL of Fief tenant
    client_id: str  # ID of Fief client
    client_secret: SecretStr  # Secret of Fief client

    class Config(PluginModel.Config):
        env_prefix = "fps_auth_fief_"
        # config can be set with environment variables, e.g.:
        # export FPS_AUTH_FIEF_BASE_URL=https://jupyverse.fief.dev
        # export FPS_AUTH_FIEF_CLIENT_ID=my_client_id
        # export FPS_AUTH_FIEF_CLIENT_SECRET=my_client_secret


def get_auth_fief_config():
    return get_config(AuthFiefConfig)


c = register_config(AuthFiefConfig)



from typing import Any, Dict, List, Optional, Tuple

from fastapi import WebSocket

from .models import User

USER = User()


def current_user(*args, **kwargs):
    async def _():
        return USER

    return _


def websocket_auth(permissions: Optional[Dict[str, List[str]]] = None):
    async def _(
        websocket: WebSocket,
    ) -> Optional[Tuple[WebSocket, Optional[Dict[str, List[str]]]]]:
        return websocket, permissions

    return _


async def update_user():
    async def _(data: Dict[str, Any]) -> User:
        global USER
        user = dict(USER)
        user.update(data)
        USER = User(**user)
        return USER

    return _



import uuid
from typing import Dict, List, Optional

from fastapi_users import schemas
from pydantic import BaseModel


class Permissions(BaseModel):
    permissions: Dict[str, List[str]]


class JupyterUser(Permissions):
    anonymous: bool = True
    username: str = ""
    name: str = ""
    display_name: str = ""
    initials: Optional[str] = None
    color: Optional[str] = None
    avatar_url: Optional[str] = None
    workspace: str = "{}"
    settings: str = "{}"


class UserRead(schemas.BaseUser[uuid.UUID], JupyterUser):
    pass


class UserCreate(schemas.BaseUserCreate, JupyterUser):
    pass


class UserUpdate(schemas.BaseUserUpdate, JupyterUser):
    pass
